"""
The Engine is the entry-point for autocron.

To use the Engine create an instance and call ``.start(filename)`` on
this instance with the database filename to use as argument. The
function ``start(filename)`` in the ``__init__.py`` module of autocron
handles this.
"""

import os
import pathlib
import signal
import subprocess
import sys
import threading
import time

from .sqlite_interface import SQLiteInterface


WORKER_MODULE_NAME = "worker.py"
WORKER_START_DELAY = 0.02


def start_subprocess(database_file):
    """
    Starts the worker process in a detached subprocess. The
    `database_file` is a string with an absolute or relative path to the
    database in use.
    """
    worker_file = pathlib.Path(__file__).parent / WORKER_MODULE_NAME
    cmd = [sys.executable, worker_file]
    if database_file:
        cmd.append(database_file)
    cwd = pathlib.Path.cwd()
    return subprocess.Popen(cmd, cwd=cwd)


class Engine:
    """
    The Engine is the entry-point for autocron. Starting the engine will
    start the monitor thread. The monitor-thread in turn starts and
    supervise the workers. (The interface argument is used for testing.)
    """
    def __init__(self, interface=None):
        # the ìnterface argument is for testing. In production this
        # argument is not provided for initialization.
        self.interface = interface if interface else SQLiteInterface()
        self.exit_event = None
        self.monitor_thread = None
        self.orig_signal_handlers = {}
        self.processes = []

    def worker_monitor(self):
        """
        Starts the worker processes and monitors that the workers are up
        and running. Restart workers if necessary. This function must
        run in a separate thread. The 'exit_event' is a
        threading.Event() instance and the `database_file` is a string
        with an absolute or relative path to the database in use. If the
        monitor receives an exit_event the function will return,
        terminating its own thread.
        """
        database_file = self.interface.db_name
        timeout = self.interface.monitor_idle_time

        for _ in range(self.interface.max_workers):
            self.processes.append(start_subprocess(database_file))
            time.sleep(WORKER_START_DELAY)

        while True:
            for process in self.processes:
                if process.poll() is not None:
                    self.interface.decrement_running_workers(process.pid)
                    self.processes.remove(process)
                    self.processes.append(start_subprocess(database_file))
                    # in case more workers need a restart:
                    time.sleep(WORKER_START_DELAY)
            if self.exit_event.wait(timeout=timeout):
                break

    def set_signal_handlers(self):
        """
        Set self._terminate() as handler for a couple of
        termination-signals and store the orinal handlers for this
        signals.
        """
        signalnums = [
            signal.SIGINT,
            signal.SIGTERM,
            signal.SIGABRT,
            signal.SIGHUP,  # availablity: Unix
        ]
        for signalnum in signalnums:
            self.orig_signal_handlers[signalnum] = signal.getsignal(signalnum)
            signal.signal(signalnum, self._terminate)

    def start(self, database_file, workers=None):
        """
        Starts the autocron worker in case autocron is active and no
        other application process has already started the workers. The
        ``database_file`` argument is a string with the name of the
        database to use (like "the_application.db") or a Path instance.
        If the name represents a relative path, the database is stored
        in the ``~/.autocron`` directory. This directory will get
        created if not already existing. If the name represents an
        absolute path, then this path will be used as is. In this case
        all directories of the path must exist.

        The ``workers`` argument takes the number of workers (as
        integer) and stores this value in the database. If the value is
        ``None`` (default) the number of workers is taken from the
        database.

        The function returns a boolean: ``True`` if workers have been
        started and ``False`` otherwise. A return value of ``False``
        does not mean that no workers are running – another application
        process may have been first on aquiring the rights to start and
        monitor the worker processes.
        """
        result = False
        self.interface.init_database(database_file)

        # don't start the engine if autocron is not active:
        if self.interface.autocron_lock:
            return result

        # autocron is active:
        # check whether start() has been called from a worker process
        # (this can happen depending on the framework architecture)
        pid = os.getpid()
        if self.interface.is_worker_pid(pid):
            # in this case the engine should not start
            return result

        # check whether the process monitors the workers,
        # but dont't start the monitor twice:
        if self.interface.acquire_monitor_lock() and not self.monitor_thread:

            # adapt number of workers if given
            if workers is not None:
                # override the already loaded value
                self.interface.max_workers = workers
                # and update the settings
                settings = self.interface.get_settings()
                settings.max_workers = workers
                self.interface.update_settings(settings)

            self.set_signal_handlers()
            self.exit_event = threading.Event()
            self.monitor_thread = threading.Thread(target=self.worker_monitor)
            self.monitor_thread.start()
            result = True

        # start the registrator thread to populate the database:
        self.interface.registrator.start()
        return result

    def stop(self):
        """
        Terminate the workers and tear-down the database. This method is
        called when the application itself terminates. It is not
        necessary to call this method directly.
        """
        if self.monitor_thread:
            # check for self.exit_event for a test-scenario.
            # in production if self.monitor_thread is not None
            # self.exit_event is also not None
            if self.exit_event:
                # terminate the monitor thread
                self.exit_event.set()
            self.monitor_thread = None

        # terminate the workers (if any) here in the main process:
        # don't do this in the monitor thread, because the monitor thread
        # may be unable to send the terminate signal to all workers before
        # the main process terminates.
        for process in self.processes:
            process.terminate()

        # stop registration and clean up the database
        self.interface.registrator.stop()
        self.interface.tear_down_database()

    def _terminate(self, signalnum, stackframe=None):
        """
        Terminate autocron by calling the engine.stop() method.
        Afterward reraise the signal again for the original
        signal-handler.
        """
        # a stackframe may be given to the signal-handler
        # which is not used here.
        # pylint: disable=unused-argument
        self.stop()
        signal.signal(signalnum, self.orig_signal_handlers[signalnum])
        signal.raise_signal(signalnum)  # requires Python >= 3.8




import pathlib
import time

import pytest

from autocron import decorators
from autocron import sqlite_interface
# from autocron import worker


TEST_DB_NAME = "test.db"


# test marker functions:
def tst_cron():
    """tst_cron_docstring"""
    pass

def tst_add(a, b):
    """tst_add_docstring"""
    return a + b

def tst_div(a, b):
    """tst_div_docstring"""
    return a / b


@pytest.fixture
def interface():
    """
    Returns a new initialised database instance.
    """
    # set class attribute to None to not return a singleton
    sqlite_interface.SQLiteInterface._instance = None
    interface = sqlite_interface.SQLiteInterface()
    interface.init_database(db_name=TEST_DB_NAME)
    # inject the interface to the decorator module
    # so the decorators access the same interface
    decorators_interface = decorators.interface
    decorators.interface = interface
    yield interface
    decorators.interface = decorators_interface
    if interface.db_name is not None:
        pathlib.Path(interface.db_name).unlink(missing_ok=True)


def test_cron_inactive(interface):
    """
    The cron decorator should return the original function but should
    not register an entry in the database. As the registrator thread is
    not running, the registration end in the queue and nothing else
    happens.
    """
    wrapper = decorators.cron()
    result = wrapper(tst_cron)
    assert result is tst_cron
    assert interface.count_tasks() == 0


def test_cron_active(interface):
    """
    The cron decorator should return the original function and should
    it save in the database.
    """
    wrapper = decorators.cron()
    result = wrapper(tst_cron)
    assert result is tst_cron
    # start registrator after execution of the decorator
    # processing the already populated queue.
    interface.registrator.start()
    time.sleep(0.1)
    assert interface.count_tasks() == 1
    interface.registrator.stop()


def test_delay_inactive(interface):
    """
    In inactive mode the delay decorator returns a result-instance with
    the function result.
    """
    interface.autocron_lock = True  # autocron is inactive
    wrapper = decorators.delay(tst_add)
    result = wrapper(40, 2)
    assert result.is_ready() is True
    assert result.has_error is False
    assert result.function_result == 42
    assert interface.count_tasks() == 0

    # raise an error during function execution:
    wrapper = decorators.delay(tst_div)
    result = wrapper(1, 0)
    assert result.is_ready() is True
    assert result.has_error is True
    assert result.function_result is None


def test_delay_active(interface):
    """
    In active mode the delay decorator returns a result-instance in
    waiting state and creates a task- and a result-entry in the
    database.
    """
    interface.registrator.start()
    assert interface.count_tasks() == 0
    assert interface.count_results() == 0
    wrapper = decorators.delay(tst_add)
    result = wrapper(40, 2)
    assert result.is_ready() is False
    time.sleep(0.1)
    assert interface.count_tasks() == 1
    assert interface.count_results() == 1
    interface.registrator.stop()


def test_delay_comment():
    """
    check for functool.wraps
    """
    wrapper = decorators.delay(tst_add)
    assert callable(wrapper) is True
    assert wrapper.__doc__ == tst_add.__doc__



"""
Scheduler for cron-tasks.

Takes a last schedule datetime-object and a crontab string and calcultes
the next time a command should get executed. The crontab format used has
five fields and a resolution of one minute
(https://en.wikipedia.org/wiki/Cron):

# ┌───────────── minute (0–59)
# │ ┌───────────── hour (0–23)
# │ │ ┌───────────── day of the month (1–31)
# │ │ │ ┌───────────── month (1–12)
# │ │ │ │ ┌───────────── day of the week (0–6) (Sunday to Saturday;
# │ │ │ │ │                                   7 is also Sunday on some systems)
# │ │ │ │ │
# │ │ │ │ │
# * * * * * <command to execute>

The fields are separated by at least a single whitespace. The fields
themselfes don't have whitespaces.

Every field can have the values: "*", "*/n" or "x,y,z".
Also a range "m-n" or a combination of list and
range like "a,b,m-n,c,p-q" is allowed.
"""

import calendar
import datetime
import re
import types


CRONTAB_PARTS = ["minutes", "hours", "days", "months", "days_of_week"]
CRONTAB_MAX_VALUES = [59, 23, 31, 12, 6]
CRONTAB_MIN_VALUES = [0, 0, 1, 1, 0]
CRONTAB_SUBSTITUTE = re.compile(r"\[|\]|\s|_")

RE_REPEAT = re.compile(r"\*/(\d+)")
RE_SEQUENCE = re.compile(r"(\d+)-(\d+)")

DAYS_PER_WEEK = 7
MAX_SCHEDULE_ITERATIONS = 10_000
MAX_SCHEDULE_ITERATIONS_ERROR_MSG = """\
max schedule iteration ({}) exceeded. Date to far in the future.
Current date: {}"""


def get_next_value(value, values):
    """
    Returns the next value from values which is larger then value or
    None if there is no larger value. Assumes the values are in sorted
    order.
    """
    for item in values:
        if item > value:
            return item
    return None


def get_numeric_sequence(pattern, min_value, max_value):
    """
    Converts a pattern to a numeric sequence:
    * -> [0..max_value] stepwidth 1
    */n -> [0..max_value] stepwidth n
    m-n -> [m..n] stepwidth 1
    m,n -> [m,n]
    a-b,m,p-q -> [a..b,m,p..q] partial stepwidth 1
    """
    # handle the * case
    if pattern == "*":
        return list(range(min_value, max_value + 1))

    # handle the */n case
    if mo := RE_REPEAT.match(pattern):
        stepwidth = int(mo.group(1))
        return list(range(min_value, max_value + 1, stepwidth))

    # handle everything else
    values = []
    for element in pattern.split(","):
        if mo := RE_SEQUENCE.match(element):
            values.extend(list(range(int(mo.group(1)), int(mo.group(2)) + 1)))
        else:
            values.append(int(element))
    return sorted(set(values))


def get_cron_parts(crontab):
    """
    Returns a SimpleNamespace object with attribute-names given in
    CRONTAB_PARTS and values provided by get_numeric_sequence() for the
    given crontab parts. Example:

    >>> cp = get_cron_parts("* 5 2-4 11 *")
    >>> cp.hours
    [5]
    >>> cp.days
    [2, 3, 4]

    The other attributes are also lists with the according values.
    """
    data = {
        name: get_numeric_sequence(item, min_value, max_value)
        for name, item, min_value, max_value in zip(
            CRONTAB_PARTS,
            crontab.split(),
            CRONTAB_MIN_VALUES,
            CRONTAB_MAX_VALUES
        )
    }
    return types.SimpleNamespace(**data)


def get_days_per_month(year=None, month=None, schedule=None):
    """
    Takes year and month and returns the number of days of the scheduled
    month. In case that schedule is given (as a datetime object), the
    year- and month-attributes from the schedule are used (shortcut
    function for better readability).
    """
    if schedule:
        year = schedule.year
        month = schedule.month
    _, days_per_month = calendar.monthrange(year, month)
    return days_per_month


def get_weekday(year=None, month=None, day=None, schedule=None):
    """
    Takes year, month and day and returns the weekday as an integer,
    where monday is 1 and sunday is 0. This is in sync with the used
    crontab but differs from the calender-module, where monday is 0 and
    sunday is 6. If schedule is given this is assumed to be a datetime
    object and in this case the other arguments are ignored.
    """
    if schedule is not None:
        year = schedule.year
        month = schedule.month
        day = schedule.day
    weekday = calendar.weekday(year, month, day) + 1
    return 0 if weekday > 6 else weekday


class CronScheduler():
    """
    Schedules a cron task.

    Usage:

    cs = CronScheduler()
    next_schedule = cs.get_next_schedule()
    """

    # pylint: disable=too-many-arguments
    def __init__(self,
                 crontab=None,
                 minutes=None,
                 hours=None,
                 days=None,
                 months=None,
                 days_of_week=None,
                 strict_mode=False):
        if not crontab:
            items = [str(item) if item else "*"
                     for item in (minutes, hours, days, months, days_of_week)]
            crontab = CRONTAB_SUBSTITUTE.sub(
                lambda mo: " " if mo.group() == "_" else "",
                "_".join(items)
            )
        self.cron_parts = get_cron_parts(crontab)
        self.strict_mode = strict_mode

    @property
    def all_weekdays_allowed(self):
        """
        Returns true if in the crontab all days of a week are allowed,
        i.e. the asterisk "*" is set.
        """
        return len(self.cron_parts.days_of_week) >= DAYS_PER_WEEK

    def get_next_schedule(self, previous_schedule=None):
        """
        Calculates the next schedule based on the current date or the
        given previous_schedule (a datetime-object). Returns a
        datetime-object.
        """
        dt = datetime.datetime
        if previous_schedule is None:
            previous_schedule = dt.now()

        year = previous_schedule.year
        month = previous_schedule.month
        day = previous_schedule.day
        hour = previous_schedule.hour
        minute = previous_schedule.minute

        minute = self.get_next_minute(minute)
        if minute is not None:
            return dt(year, month, day, hour, minute)

        minute = self.cron_parts.minutes[0]  # get first minute
        hour = self.get_next_hour(hour)
        if hour is not None:
            return dt(year, month, day, hour, minute)

        hour = self.cron_parts.hours[0]  # get first hour
        day = self.get_next_day(year, month, day)
        if day is not None:
            return dt(year, month, day, hour, minute)

        for counter in range(MAX_SCHEDULE_ITERATIONS):
            month = self.get_next_month(month)
            if month:
                day = self.get_first_day(year, month)
                if day is not None:
                    return dt(year, month, day, hour, minute)
                continue
            year += 1
            month = self.cron_parts.months[0]  # get first month
            day = self.get_first_day(year, month)
            if day is not None:
                return dt(year, month, day, hour, minute)

        # not returning from inside the loop is a potential
        # endless loop. So after MAX_SCHEDULE_ITERATIONS there
        # is a hard break here:
        schedule = dt(year, month, day, hour, minute)
        msg = MAX_SCHEDULE_ITERATIONS_ERROR_MSG.format(counter, schedule)
        raise ValueError(msg)

    def get_next_minute(self, minute):
        """
        Returns the next minute of configured minutes. Returns None if
        there is no next minute after the given one.
        """
        return get_next_value(minute, self.cron_parts.minutes)

    def get_next_hour(self, hour):
        """
        Returns the next hour of configured hours. Returns None if
        there is no next hour after the given one.
        """
        return get_next_value(hour, self.cron_parts.hours)

    def get_first_day(self, year, month):
        """
        Wrapper for get_next_day with day=0 to get the first allowed day
        of a month or None, if the month has no allwed days.
        """
        return self.get_next_day(year, month, day=0)

    def get_next_day(self, year, month, day):
        """
        Returns the next allowed day after the given day. Returns the
        day or None in case that there is no follow up day for the month
        and year.
        """
        # pylint:disable=too-many-return-statements
        next_day = get_next_value(day, self.cron_parts.days)

        if self.all_weekdays_allowed:
            # strict_mode doesn't matter
            if next_day is not None:
                if next_day <= get_days_per_month(year, month):
                    return next_day
            return None

        if self.strict_mode:
            if next_day is None:
                return None
            # check whether next_day is an allowed day of the
            # current month.
            if next_day > get_days_per_month(year, month):
                return None
            # the day must match one of the allowed weekdays
            weekday_of_next_day = get_weekday(year, month, next_day)
            if weekday_of_next_day in self.cron_parts.days_of_week:
                return next_day
            # recursion is save because there are just a few days per month
            return self.get_next_day(year, month, next_day)

        # no strict mode but days_of_week are defined:
        # the next day and the next allowed weekday are valid days,
        # return the one that comes first

        # `day` could be zero to get the first day of the month.
        # this is necessary for the first get_next_value() call
        # but will trigger a ValueError in the calendar module.
        # in this case set day to 1:

        if day == 0:
            day = 1
        weekday_of_day = get_weekday(year, month, day)
        next_weekday = get_next_value(
            weekday_of_day,
            self.cron_parts.days_of_week
        )
        if next_weekday is None:
            next_weekday = self.cron_parts.days_of_week[0] + DAYS_PER_WEEK
        delta = next_weekday - weekday_of_day
        next_weekday_day = day + delta

        if next_day is None:
            next_day = next_weekday_day
        else:
            next_day = min(next_day, next_weekday_day)
        if next_day <= get_days_per_month(year, month):
            return next_day
        return None

    def get_next_month(self, month):
        """
        Returns the next month of configured months. Returns None if
        there is no next month after the given one.
        """
        return get_next_value(month, self.cron_parts.months)



"""
admin.py

Administration tool to access the database.
"""

import argparse

from autocron.sqlite_interface import (
    SQLiteInterface,
    SETTINGS_DEFAULT_DATA,
)

interface = SQLiteInterface()


PGM_NAME = "autocron command line tool"
PGM_DESCRIPTION = """
    Provides access to the autocron database to change settings
    an get information about the running workers.
"""


class Admin:
    """Collection of methods for administration-tasks.
    """

    def __init__(self, db_name):
        self.interface = SQLiteInterface()
        self.interface.init_database(db_name)

    def show_info(self):
        """Tabular view of the settings.
        """
        settings = self.interface.get_settings()
        column_width = len(max(settings.columns, key=len))
        database = "database"
        print(f"\n{database:<{column_width}}: '{self.interface.db_name}'")
        print(settings)
        tasks = self.interface.count_tasks()
        results = self.interface.count_results()
        for name, value in zip(("tasks", "results"), (tasks, results)):
            print(f"{name:<{column_width}}: {value}")
        print()

    def set_max_workers(self, workers):
        """Set the number of workers.
        """
        settings = self.interface.get_settings()
        settings.max_workers = workers
        self.interface.update_settings(settings)
        print(f"set max_workers to {workers}")

    def set_autocron_lock(self, flag):
        """Set the autocron_lock flag.
        """
        settings = self.interface.get_settings()
        settings.autocron_lock = convert_flag(flag)
        self.interface.update_settings(settings)
        print(f"set autocron lock to {flag}")

    def set_monitor_lock(self, flag):
        """Set the monitor_lock flag.
        """
        settings = self.interface.get_settings()
        settings.monitor_lock = convert_flag(flag)
        self.interface.update_settings(settings)
        print(f"set monitor lock to {flag}")

    def set_worker_idle_time(self, idle_time):
        """
        Set the idle time of the worker in seconds. This is the time the
        worker sleeps when no new tasks are on due.
        """
        settings = self.interface.get_settings()
        settings.worker_idle_time = idle_time
        self.interface.update_settings(settings)
        print(f"set worker idle time to {idle_time} seconds")

    def set_monitor_idle_time(self, idle_time):
        """
        Set the idle time of the worker in seconds. This is the time the
        worker sleeps when no new tasks are on due.
        """
        settings = self.interface.get_settings()
        settings.monitor_idle_time = idle_time
        self.interface.update_settings(settings)
        print(f"set monitor idle time to {idle_time} seconds")

    def set_result_ttl(self, ttl):
        """
        Set the result time to life in seconds. ttl is an integer. This is
        the timespan a result will get stored in the database.
        """
        settings = self.interface.get_settings()
        settings.result_ttl = ttl
        self.interface.update_settings(settings)
        print(f"Set result-ttl to {ttl} seconds")

    def set_defaults(self):
        """Reset all settings to the default values.
        """
        settings = self.interface.get_settings()
        settings.__dict__.update(SETTINGS_DEFAULT_DATA)
        self.interface.update_settings(settings)
        print("\nautocron reset to default data:")
        self.show_info()

    def delete_database(self):
        """
        Delete the sqlite-database. A new one will get created on the
        next start of the admin-tool or autocron.
        """
        answer = input("sure to delete the current database? [y/n]: ")
        if answer.lower() == 'y':
            self.interface.db_name.unlink(missing_ok=True)
            self.interface.db_name = None
        else:
            print("abort command")


def convert_flag(flag):
    """
    flag can be a string with 'true'', 'false' or 'on'', 'off''.
    Returns a boolean.
    """
    flag = flag.lower()
    return flag in {"true", "on"}


def print_usage():
    """Print program-description and hint how to get help."""
    print(f"\n{PGM_NAME}\n{PGM_DESCRIPTION}\nuse option -h for help.\n\n")


def get_command_line_arguments():
    """Get the command line arguments.
    """
    parser = argparse.ArgumentParser(
        prog=PGM_NAME,
        description=PGM_DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "database",
        help="database name or path."
    )
    parser.add_argument(
        "-i", "--info",
        action="store_true",
        help="provide information about the settings and process data."
    )
    parser.add_argument(
        "--set-max-workers",
        dest="max_workers",
        type=int,
        help="set number of worker processes."
    )
    parser.add_argument(
        "--set-autocron-lock",
        dest="autocron_lock",
        help="set autocron lock flag: [true|false or on|off]."
    )
    parser.add_argument(
        "--set-monitor-lock",
        dest="monitor_lock",
        help="set monitor lock flag: [true|false or on|off]."
    )
    parser.add_argument(
        "--set-worker-idle-time",
        dest="worker_idle_time",
        type=int,
        help="set worker idle time in seconds [set to 0 for auto-calculation]."
    )
    parser.add_argument(
        "--set-monitor-idle-time",
        dest="monitor_idle_time",
        type=int,
        help="set monitor idle time in seconds."
    )
    parser.add_argument(
        "--set-result-ttl",
        dest="result_ttl",
        type=int,
        help="set result time to life in seconds."
    )
    parser.add_argument(
        "--set-defaults",
        dest="set_defaults",
        action="store_true",
        help="reset to default settings."
    )
    parser.add_argument(
        "--delete-database",
        dest="delete_database",
        action="store_true",
        help="delete the current database."
    )
    return parser.parse_args()


def main():
    """entry point.
    """
    args = get_command_line_arguments()
    admin = Admin(args.database)
    if args.info:
        admin.show_info()
    elif args.max_workers:
        admin.set_max_workers(args.max_workers)
    elif args.autocron_lock:
        admin.set_autocron_lock(args.autocron_lock)
    elif args.monitor_lock:
        admin.set_monitor_lock(args.monitor_lock)
    elif args.worker_idle_time:
        admin.set_worker_idle_time(args.worker_idle_time)
    elif args.monitor_idle_time:
        admin.set_monitor_idle_time(args.monitor_idle_time)
    elif args.result_ttl:
        admin.set_result_ttl(args.result_ttl)
    elif args.set_defaults:
        admin.set_defaults()
    elif args.delete_database:
        admin.delete_database()
    else:
        print_usage()


if __name__ == "__main__":
    main()



"""
autocron:

simple asynchronous background task handling with no dependencies beside
the standard-library.
"""

from .decorators import (
    cron,
    delay,
)
from .engine import Engine
from .sqlite_interface import SQLiteInterface


__all__ = ["cron", "delay", "start", "stop"]
__version__ = "0.9.5"

_engine = Engine()
_interface = SQLiteInterface()


def start(database_file, workers=None):
    """
    Call this from the framework of choice to explicitly activate
    autocron. ``database_file`` is a string with the file-name of the
    database. The file gets stored in the ``~.autocron/`` directory. If
    this directory does not exist, it will get created. The file-name can
    also be an absolute path so the file will get stored elsewere. In
    this case all directories in the path must exist.

    With ``workers`` the number of worker is set and stored in the
    database. If the value is ``None`` (default) the number of workers
    are read from the database.
    """
    _engine.start(database_file=database_file, workers=workers)


def stop():
    """
    Stops autocron explicitly. On receiving a termination-signal
    autocron invokes a shutdown sequence to stop the workers, so calling
    ``stop`` is normalwise not required.
    """
    _engine.stop()



"""
test_sqlite_interface.py
"""

import datetime
import pathlib
import uuid

import pytest

from autocron import sqlite_interface

from autocron.sqlite_interface import (
    SETTINGS_DEFAULT_WORKERS,
    TASK_STATUS_WAITING,
    TASK_STATUS_PROCESSING,
    Connection,
    Settings,
    Result,
    Task
)


TEST_DB_NAME = "test.db"


def tst_function(*args, **kwargs):
    return args, kwargs


def tst_cron_function():
    return None


def tst_add_function(a, b):
    return a + b


@pytest.fixture
def raw_interface():
    """
    Returns a new uninitialised database instance.
    """
    # set class attribute to None to not return a singleton
    sqlite_interface.SQLiteInterface._instance = None
    interface = sqlite_interface.SQLiteInterface()
    yield interface
    if interface.db_name is not None:
        pathlib.Path(interface.db_name).unlink(missing_ok=True)


@pytest.fixture
def interface():
    """
    Returns a new initialised database instance.
    """
    # set class attribute to None to not return a singleton
    sqlite_interface.SQLiteInterface._instance = None
    interface = sqlite_interface.SQLiteInterface()
    interface.init_database(db_name=TEST_DB_NAME)
    yield interface
    if interface.db_name is not None:
        pathlib.Path(interface.db_name).unlink(missing_ok=True)


@pytest.mark.parametrize(
    "db_name, parent_dir", [
        ("db_name.db", ".autocron"),
        ("path/db_name.db", ".autocron"),
        ("/path/db_name.db", "path")
    ]
)
def test_storage_location(db_name, parent_dir, raw_interface):
    """
    Test for storage location: a relative path gets stored to ~.autocron
    and an absolute path as is.
    """
    raw_interface.db_name = db_name
    parent = raw_interface.db_name.parent
    assert parent_dir == parent.stem


def test_init_database(interface):
    """
    Initialize with and without settings.
    """
    # init_database has created the db and set the default settings
    # check for the settings entry in the database
    with Connection(interface.db_name) as conn:
        rows = Settings.count_rows(conn)
        assert rows == 1

    settings = interface.get_settings()
    assert settings.autocron_lock is False
    assert settings.monitor_lock is False


def test_update_settings(interface):
    """
    Test the .update() method on Model.
    """
    max_workers = 2
    with Connection(interface.db_name) as conn:
        settings = Settings.select_all(conn)[0]
        assert settings.max_workers == SETTINGS_DEFAULT_WORKERS

        # check this because otherwise the test makes no sense:
        assert max_workers != SETTINGS_DEFAULT_WORKERS
        settings.max_workers = max_workers
        settings.update()

    with Connection(interface.db_name) as conn:
        settings = Settings.select_all(conn)[0]
        assert settings.max_workers == max_workers


def test_acquire_monitor_lock(interface):
    """Test to set the monitor_lock flag.
    """
    # after initialization the monito_lock flag is False,
    # therefor acquisition is possible:
    settings = interface.get_settings()
    assert settings.monitor_lock is False
    result = interface.acquire_monitor_lock()
    assert result is True

    # monitor_lock has been set to True:
    settings = interface.get_settings()
    assert settings.monitor_lock is True

    # no acquisition is no longer possible:
    result = interface.acquire_monitor_lock()
    assert result is False

    # and the flag has not changed:
    settings = interface.get_settings()
    assert settings.monitor_lock is True


def test_crud_task(interface):
    """
    Make the crud test on the model using a task.
    """
    # C: create with attributes
    schedule = datetime.datetime(2000, 1, 1)
    args = (3.14, "test", 42)
    kwargs = {"pi": 3.14159, "a": 42, 10: "b"}
    with Connection(interface.db_name) as conn:
        task = Task(
            connection=conn, func=tst_function, schedule=schedule,
            args=args, kwargs=kwargs
        )
        task.store()

    # after storage the task should have a rowid attribute
    rowid = task.rowid

    # R: read by rowid and check for correct attribute-reading
    with Connection(interface.db_name) as conn:
        task = Task.select(conn, rowid=rowid)
    assert task.schedule == schedule
    assert task.args == args
    assert task.kwargs == kwargs
    assert task.crontab == ""
    assert task.uuid == ""
    assert task.function_module == tst_function.__module__
    assert task.function_name == tst_function.__name__

    # U: update a selected attribute:
    now = datetime.datetime.now()
    task.schedule = now
    with Connection(interface.db_name) as conn:
        task.connection = conn
        task.update()
    with Connection(interface.db_name) as conn:
        task = Task.select(conn, rowid=rowid)
    assert task.schedule == now

    # D: delete the task:
    with Connection(interface.db_name) as conn:
        task.connection = conn
        task.delete()
    with Connection(interface.db_name) as conn:
        assert Task.count_rows(conn) == 0


def test_delete_task_via_interface(interface):
    """Delete a task via an interface-method.
    """
    with Connection(interface.db_name) as conn:
        task = Task(conn, func=tst_function)
        task.store()
        assert Task.count_rows(conn) == 1

    # invalidate the connection (which already closed from the context):
    task.connection = None

    # delete the task and check whether it has worked:
    interface.delete_task(task)
    with Connection(interface.db_name) as conn:
        assert Task.count_rows(conn) == 0


def test_register_cron_task(interface):
    """
    Register a cron task should create a task entry but no result
    entry. Also crontasks should not get registered twice.
    """
    interface.register_task(tst_cron_function, crontab="*")
    with Connection(interface.db_name) as conn:
        assert Task.count_rows(conn) == 1
        assert Result.count_rows(conn) == 0

    # try to register again:
    interface.register_task(tst_cron_function, crontab="*")
    with Connection(interface.db_name) as conn:
        assert Task.count_rows(conn) == 1

    # registration of another crontask should work:
    interface.register_task(tst_function, crontab="*")
    with Connection(interface.db_name) as conn:
        assert Task.count_rows(conn) == 2

    # registration of a delayed function registered already as
    # a crontask should also work. (this is a theoretical case.)
    interface.register_task(tst_function)
    with Connection(interface.db_name) as conn:
        assert Task.count_rows(conn) == 3


def test_register_delayed_task(interface):
    """
    Register a delayed task should create a task entry and also a result
    entry.
    """
    interface.register_task(tst_cron_function, uuid="*")
    with Connection(interface.db_name) as conn:
        assert Task.count_rows(conn) == 1
        assert Result.count_rows(conn) == 1


def test_get_next_task(interface):
    """
    Test to just return a task on due.
    """
    now = datetime.datetime.now()
    delta = datetime.timedelta(hours=1)

    # add a task that is not on due:
    with Connection(interface.db_name) as conn:
        task = Task(connection=conn, func=tst_function, schedule=now+delta)
        task.store()
    next_task = interface.get_next_task()
    assert next_task is None

    # add a task that is on due:
    with Connection(interface.db_name) as conn:
        task = Task(connection=conn, func=tst_add_function, schedule=now-delta)
        task.store()
    next_task = interface.get_next_task()
    assert next_task is not None
    assert next_task.function_name == tst_add_function.__name__


def test_get_next_task_priority(interface):
    """
    Returns the next task on due and crontasks first.
    """
    delta = datetime.timedelta(hours=1)
    now = datetime.datetime.now()
    interface.register_task(
        tst_cron_function, crontab="*", schedule=now - 2 * delta)
    interface.register_task(
        tst_function, uuid="*", schedule=now - delta)
    interface.register_task(
        tst_add_function, crontab="*", schedule=now + delta)

    # get_next_task should return at first the tst_cron_function
    # and then the tst_function.
    # The tst_add_function should not get returned:

    task = interface.get_next_task()
    assert task.function_name == tst_cron_function.__name__

    task = interface.get_next_task()
    assert task.function_name == tst_function.__name__

    task = interface.get_next_task()
    assert task is None


def test_get_task_by_function_name(interface):
    """Store two tasks and get one by name.
    """
    with Connection(interface.db_name) as conn:
        Task(connection=conn, func=tst_function).store()
        Task(connection=conn, func=tst_cron_function).store()
    with Connection(interface.db_name) as conn:
        assert Task.count_rows(conn) == 2
        task = Task.get_by_function_name(tst_cron_function, conn)
        assert task is not None
        assert task.function_module == tst_cron_function.__module__
        assert task.function_name == tst_cron_function.__name__


def test_task_is_ready(interface):
    """
    Test for correct refresh of a Result instance on calling 'is_ready()'.
    """
    with Connection(interface.db_name) as conn:
        uuid_ = uuid.uuid4().hex
        result = Result(conn, func=tst_function, uuid=uuid_)
        result.store()
    assert result.uuid == uuid_
    assert result.is_ready() is False

    # update and result and check for changes:
    answer = 42
    interface.update_result(uuid=uuid_, result=answer)
    assert result.is_ready() is True
    assert result.has_error is False
    assert result.result == answer



def test_delete_outdated_results(interface):
    """
    Test to store and delete results.
    """
    delta = datetime.timedelta(hours=1)
    now = datetime.datetime.now()
    with Connection(interface.db_name) as conn:
        outdated = Result(conn, func=tst_function, uuid=uuid.uuid4().hex)
        outdated.store()
        also_outdated = Result(conn, func=tst_function, uuid=uuid.uuid4().hex)
        also_outdated.store()
        not_outdated = Result(conn, func=tst_function, uuid=uuid.uuid4().hex)
        not_outdated.store()
    assert interface.count_results() == 3

    # nothing will happen because no result is outdated
    interface.delete_outdated_results()
    assert interface.count_results() == 3

    # now update the results as it will happen by the workers:
    interface.update_result(uuid=not_outdated.uuid)
    interface.update_result(uuid=outdated.uuid, ttl=now-delta)
    interface.update_result(uuid=also_outdated.uuid, ttl=now-delta)

    assert interface.count_results() == 3  # as before
    interface.delete_outdated_results()
    assert interface.count_results() == 1  # but now two have been deleted


def test_increment_running_workers(interface):
    """
    Check to increment the list of running workers.
    """
    # add the first worker:
    first_pid = 123
    running_workers = 1
    interface.increment_running_workers(pid=first_pid)

    # check for correct setup
    with Connection(interface.db_name) as conn:
        settings = Settings.read(connection=conn)
    assert settings.worker_pids == str(first_pid)
    assert settings.running_workers == running_workers

    # provide an additional pid and call increment
    new_pid = 42
    interface.increment_running_workers(pid=new_pid)

    # check for updated settings:
    worker_pids = f"{first_pid},{new_pid}"
    running_workers += 1
    with Connection(interface.db_name) as conn:
        settings = Settings.read(connection=conn)
    assert settings.worker_pids == worker_pids
    assert settings.running_workers == running_workers


def test_decrement_running_workers(interface):
    """
    Check to decrement the list of running workers.
    """
    # setting up three running workers:
    worker_pids = "123,456,789"
    running_workers = 3
    with Connection(interface.db_name) as conn:
        settings = Settings.read(connection=conn)
        settings.running_workers=running_workers
        settings.worker_pids=worker_pids
        settings.update()

    # do the modification: delete pid 456
    pid = 456
    interface.decrement_running_workers(pid=pid)
    with Connection(interface.db_name) as conn:
        settings = Settings.read(connection=conn)
    assert settings.worker_pids == "123,789"
    assert settings.running_workers == 2

    # delete a non-existing pid should have no effect:
    pid = 42
    interface.decrement_running_workers(pid=pid)
    with Connection(interface.db_name) as conn:
        settings = Settings.read(connection=conn)
    assert settings.worker_pids == "123,789"
    assert settings.running_workers == 2

    # remove the remaining two pids:
    interface.decrement_running_workers(pid="123")
    interface.decrement_running_workers(pid="789")
    with Connection(interface.db_name) as conn:
        settings = Settings.read(connection=conn)
    assert settings.worker_pids == ""
    assert settings.running_workers == 0

    # don't go below zero for the running_workers counter
    # (should be covered by ignoring unknown pids):
    interface.decrement_running_workers(pid="23")
    with Connection(interface.db_name) as conn:
        settings = Settings.read(connection=conn)
    assert settings.worker_pids == ""
    assert settings.running_workers == 0


@pytest.mark.parametrize(
    "pids, pid, expected_result", [
        ("123,456,789,1012,1024,300,4578", 456, True),
        ("123,456,789,1012,1024,300,4578", 301, False),
        ("", 987652, False),
    ]
)
def test_is_worker_pid(pids, pid, expected_result, interface):
    with Connection(interface.db_name) as conn:
        settings = Settings.read(connection=conn)
        settings.worker_pids = pids
        settings.update()
    result = interface.is_worker_pid(pid)
    assert result == expected_result




"""
Implementation of the decorators ``cron`` and ``delay`` for running
recurring task and to delegate long running tasks to a background
process.
"""

import functools
import uuid

from .schedule import CronScheduler
from .sqlite_interface import (
    TASK_STATUS_WAITING,
    TASK_STATUS_READY,
    TASK_STATUS_ERROR,
    Result,
    SQLiteInterface
)


# default: run every minute:
DEFAULT_CRONTAB = "* * * * *"

interface = SQLiteInterface()


# pylint: disable=too-many-arguments
def cron(crontab=None,
         minutes=None,
         hours=None,
         days=None,
         months=None,
         days_of_week=None):
    """
    Decorator for a cronjob. Functions running cronjobs should not get
    called from the main program and therefore don't get arguments.
    Example usage for a cronjob to run every hour, at the beginning of
    the hour:

    >>> @cron("0 * * * *")
    >>> def some_callable():
    >>>    # do periodic stuff here ...

    The decorator can take a couple of arguments but if just the first
    argument `crontab` is given then all other arguments are ignored. To
    use the other arguments instead, provide them all as
    keyword-arguments. If no arguments are given the default-crontab
    ``(* * * * *)`` is used to execute a task every minute.

    :crontab:
        a string representing a valid crontab. See:
        `https://en.wikipedia.org/wiki/Cron#CRON_expression
        <https://en.wikipedia.org/wiki/Cron#CRON_expression>`_ with the
        restriction that only integers and the special signs (* , -) are
        allowed. Some examples ::

            The order of arguments is:
            'minutes hours dow months dom'

            '* * * * *':       runs every minute
                               (same as @periodic_task(seconds=60))
            '15,30 7 * * *':   runs every day at 7:15 and 7:30
            '* 9 0 4,7 10-15': runs at 9:00 every monday and
                               from the 10th to the 15th of a month
                               but only in April and July.


    :minutes:
        list of minutes during an hour when the task should run. Valid
        entries are integers in the range 0-59. Defaults to None which
        is the same as ``*`` in a crontab, meaning that the task gets
        executed every minute.

    :hours:
        list of hours during a day when the task should run. Valid
        entries are integers in the range 0-23. Defaults to None which
        is the same as ``*`` in a crontab, meaning that the task gets
        executed every hour.

    :days:
        list of days in an month the task should run. Valid entries are
        integers in the range 1-31. Defaults to None which is the same
        as ``*`` in a crontab, meaning that the task gets executed every
        day.

    :months:
        list of month during a year when the task should run. Valid
        entries are integers in the range 1-12. Defaults to None which
        is the same as ``*`` in a crontab, meaning that the task gets
        executed every month.

    :days_of_week:
        days of week. A list of integers from 0 to 6 with Monday as 0.
        The task runs only on the given weekdays. Defaults to None which
        is the same as ``*`` in a crontab, meaning that the task gets
        executed every day of the week.

    If neither *days* nor *days_of_week* are given, then the task will
    run every day of a month. If one of both is set, then the given
    restrictions apply. If both are set, then the allowed days
    complement each other.
    """
    # set crontab to default if no other arguments are given:
    if not any(locals().values()):
        crontab = DEFAULT_CRONTAB

    def wrapper(func):
        # send the function to the registerer. The contas will get registered
        # when autocron starts. If autocron is not active nothing bad happens.
        scheduler = CronScheduler(
            minutes=minutes,
            hours=hours,
            days=days,
            months=months,
            days_of_week=days_of_week,
            crontab=crontab
        )
        schedule = scheduler.get_next_schedule()
        interface.registrator.register(
            func,
            schedule=schedule,
            crontab=crontab
        )
        return func
    return wrapper


def delay(func):
    """
    Decorator for a delayed task. Apply this as:

    >>> @delay
    >>> def sendmail(recipient, message):
    >>>     # code goes here ...

    The decorator does not take any arguments. Calling ``sendmail()``
    will return from the call immediately and this callable will get
    executed later in another process.

    The decorated function will return a Result-instance. If autocron is
    active the result will be in waiting mode and may not be in the
    database because the dataset gets created in a separate thread. If
    autocron is not active, the result-instance will be in ready- or
    error-mode, depending on the function call.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # the wrapper will not get called during import time.
        # at runtime the database is initialized and it is safe
        # to check the settings:
        if not interface.accept_registrations:
            # this is the case when the decorated function gets called
            # in a worker process. In this case the wrapper returns the
            # result from the function call and not a Result instance,
            # because the Result instance is already existing and just
            # updated by the worker.
            # (also the error handling is done by the worker.)
            return func(*args, **kwargs)

        # in the 'main' process autocron may be active or not:
        if interface.autocron_lock:
            # inactive: call the function and return a Result-instance
            # in ready- or error-state:
            try:
                function_result = func(*args, **kwargs)
            except Exception as err:  # pylint: disable=broad-exception-caught
                error_message = str(err)
                status = TASK_STATUS_ERROR
                function_result = None
            else:
                error_message = ""
                status = TASK_STATUS_READY
            result = Result.from_registration(
                func, args, kwargs, status=status,
                function_result=function_result,
                error_message=error_message
            )
        else:
            # active: register in task_queue and return a Result-instance
            # in waiting-state:
            uuid_ = uuid.uuid4().hex
            interface.registrator.register(
                func, args=args, kwargs=kwargs, uuid=uuid_
            )
            result = Result.from_registration(
                func, args, kwargs, uuid=uuid_, status=TASK_STATUS_WAITING
            )
        return result
    return wrapper



"""
The sqlite interface.

Data models, sql-definition and access routines.
"""

# some structures are larger than pylint likes:
# pylint: disable=too-many-arguments
# pylint: disable=too-many-instance-attributes


import datetime
import functools
import pathlib
import pickle
import queue
import sqlite3
import threading
import time


DEFAULT_STORAGE = ".autocron"
REGISTER_BACKGROUND_TASK_TIMEOUT = 2.0

SQLITE_OPERATIONAL_ERROR_RETRIES = 100
SQLITE_OPERATIONAL_ERROR_DELAY = 0.01
SQLITE_DELAY_INCREMENT_STEPS = 20
SQLITE_DELAY_INCREMENT_FACTOR = 1.5

SQLITE_EXCLUSIVE_ACCESS = "BEGIN EXCLUSIVE"

SETTINGS_DEFAULT_WORKERS = 1
SETTINGS_DEFAULT_RUNNING_WORKERS = 0
SETTINGS_DEFAULT_MONITOR_LOCK = False
SETTINGS_DEFAULT_AUTOCRON_LOCK = False
SETTINGS_DEFAULT_MONITOR_IDLE_TIME = 5  # seconds
SETTINGS_DEFAULT_WORKER_IDLE_TIME = 0  # 0 seconds means auto idle time
SETTINGS_DEFAULT_WORKER_PIDS = ""
SETTINGS_DEFAULT_RESULT_TTL = 1800  # Storage time (time to live) in seconds

SETTINGS_DEFAULT_DATA = {
    "max_workers": SETTINGS_DEFAULT_WORKERS,
    "running_workers": SETTINGS_DEFAULT_RUNNING_WORKERS,
    "monitor_lock": SETTINGS_DEFAULT_MONITOR_LOCK,
    "autocron_lock": SETTINGS_DEFAULT_AUTOCRON_LOCK,
    "monitor_idle_time": SETTINGS_DEFAULT_MONITOR_IDLE_TIME,
    "worker_idle_time": SETTINGS_DEFAULT_WORKER_IDLE_TIME,
    "worker_pids": SETTINGS_DEFAULT_WORKER_PIDS,
    "result_ttl": SETTINGS_DEFAULT_RESULT_TTL
}

BOOLEAN_SETTINGS = ["monitor_lock", "autocron_lock"]

# Status codes used for task-status the result-entries:
TASK_STATUS_WAITING = 1
TASK_STATUS_PROCESSING = 2
TASK_STATUS_READY = 3
TASK_STATUS_ERROR = 4
TASK_STATUS_UNAVAILABLE = 5


# sqlite3 default adapters and converters deprecated as of Python 3.12:
def datetime_adapter(value):
    """
    Gets a python datetime-instance and returns an ISO 8601 formated
    string for sqlite3 storage.
    """
    return value.isoformat()


def datetime_converter(value):
    """
    Gets an ISO 8601 formated byte-string (from sqlite3) and returns a
    python datetime datatype.
    """
    return datetime.datetime.fromisoformat(value.decode())


sqlite3.register_adapter(datetime.datetime, datetime_adapter)
sqlite3.register_converter("datetime", datetime_converter)


# sqlite3: decorator for SQLiteInterface-methods accessing the database
def db_access(function):
    """
    Access decorator. Repeats the decorated function several times in
    case the database is locked because of write- or exclusive-access.
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        """
        Repeat the wrapped function call in case of an OperationalError.
        If this fails for SQLITE_MAX_RETRY_LIMIT times, the original
        error is raised.
        """
        message = ""
        delay = SQLITE_OPERATIONAL_ERROR_DELAY
        for retry_num in range(SQLITE_OPERATIONAL_ERROR_RETRIES):
            try:
                return function(*args, **kwargs)
            except sqlite3.OperationalError as err:
                message = str(err)
                time.sleep(delay)
            if not retry_num % SQLITE_DELAY_INCREMENT_STEPS:
                delay *= SQLITE_DELAY_INCREMENT_FACTOR
        raise sqlite3.OperationalError(message)
    return wrapper


class Model:
    """
    Base model for the table classes for common sql-creation and access
    methods.
    """

    # class attributes to redefine in subclasses:
    table_name = ""
    columns = {}

    def __init__(self, connection=None):
        self.connection = connection
        self.rowid = None

    def store(self):
        """
        Store a new row. data is a dictionary with all column data.
        After storage the instance-attribute `rowid` will be set.
        """
        columns = ",".join(f":{name}" for name in self.columns)
        sql = f"""INSERT INTO {self.table_name} VALUES ({columns})
                  RETURNING rowid"""
        cursor = self.connection.run(sql, self.__dict__)
        result = cursor.fetchone()
        # result is a tuple representing the RETURNING values
        # from the sql command. In this case it is a tuple with
        # a single entry holding the new created rowid:
        self.rowid = result[0]
        return self.rowid

    def update(self):
        """Make the current set of attributes persistent.
        """
        columns = ",".join(f"{name} = :{name}" for name in self.columns)
        sql = f"""UPDATE {self.table_name} SET {columns}
                  WHERE rowid == :rowid"""
        self.connection.run(sql, self.__dict__)

    def delete(self):
        """Delete this instance by the rowid.
        """
        sql = f"DELETE FROM {self.table_name} WHERE rowid == {self.rowid}"
        self.connection.run(sql)

    @classmethod
    def _get_sql_select(cls):
        """Helper function for the select-classmethods.
        """
        columns = list(cls.columns.keys())
        columns.append("rowid")
        columns = ",".join(columns)
        return f"SELECT {columns} FROM {cls.table_name}"

    @classmethod
    def select(cls, connection, rowid=None, sql=None, data=None):
        """
        Returns an instance depending on the arguments or None if no
        entry matches. If just `rowid` is given makes a lookup by the
        `rowid`. Otherwise `sql` and `data` must be given to make
        specific selection.
        """
        if rowid is not None:
            # simple select by rowid
            sql = cls._get_sql_select()
            sql = f"{sql} WHERE rowid == :rowid"
            data = {"rowid": rowid}
        cursor = connection.run(sql, data)
        cursor.row_factory = getattr(cls, "row_factory", None)
        data = cursor.fetchone()
        instance = None
        if data:
            instance = cls(connection)
            instance.__dict__.update(data)
        return instance

    @classmethod
    def select_all(cls, connection):
        """
        Select all entries from a table and returns a list of
        model-instances. If there is no entry in the table an empty list
        is returned.
        """
        sql = cls._get_sql_select()
        cursor = connection.run(sql)
        cursor.row_factory = getattr(cls, "row_factory", None)
        data_set = cursor.fetchall()
        instances = []
        for data in data_set:
            instance = cls(connection)
            instance.__dict__.update(data)
            instances.append(instance)
        return instances

    @classmethod
    def create_table(cls, connection):
        """Create the database table for the model if not already existing.
        """
        columns = ",".join(
            f"{field} {type}" for field, type in cls.columns.items()
        )
        connection.run(
            f"CREATE TABLE IF NOT EXISTS {cls.table_name}({columns})"
        )

    @classmethod
    def count_rows(cls, connection):
        """Return the number of rows in the table.
        """
        cursor = connection.run(f"SELECT COUNT(*) FROM {cls.table_name}")
        return cursor.fetchone()[0]


class Task(Model):
    """Model to store functions for later execution.
    """

    table_name = "task"
    columns = {
        "uuid": "TEXT",
        "schedule": "datetime",
        "status": "INTEGER",
        "crontab": "TEXT",
        "function_module": "TEXT",
        "function_name": "TEXT",
        "function_arguments": "BLOB"
    }

    def __init__(
        self,
        connection=None,
        func=None,
        args=(),
        kwargs=None,
        uuid="",
        crontab="",
        schedule=None,
        status=TASK_STATUS_WAITING
    ):
        super().__init__(connection=connection)
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.uuid = uuid
        self.crontab = crontab
        self.schedule = schedule
        self.status = status
        self.function_module = None
        self.function_name = None
        self.function_arguments = None

    def store(self):
        """
        Store a new task in the database. Returns the rowid of the new dataset.
        """
        if self.schedule is None:
            self.schedule = datetime.datetime.now()
        if self.func:
            self.function_module = self.func.__module__
            self.function_name = self.func.__name__
        else:
            self.function_module = ""
            self.function_name = ""
        self.function_arguments = pickle.dumps((self.args, self.kwargs))
        super().store()

    def update(self):
        """Make the current state of attributes persistent.
        """
        # function arguments may have changed:
        self.function_arguments = pickle.dumps((self.args, self.kwargs))
        super().update()

    @classmethod
    def _get_next_task_sql_and_data(cls):
        """Helper method for next_task and next_cron_task.
        """
        sql = cls._get_sql_select()
        sql = f"""{sql} WHERE schedule <= :schedule
                  AND status == {TASK_STATUS_WAITING}"""
        data = {"schedule": datetime.datetime.now()}
        return sql, data

    @classmethod
    def get_by_function_name(cls, function, connection):
        """
        Return a task instance identified by the given function or None
        if no task is found.
        """
        sql = cls._get_sql_select()
        sql = f"""{sql} WHERE function_module == :function_module
                        AND function_name == :function_name"""
        data = {
            "function_module": function.__module__,
            "function_name": function.__name__
        }
        return cls.select(connection=connection, sql=sql, data=data)

    @classmethod
    def next_task(cls, connection):
        """Returns a task instance which is on due.
        """
        sql, data = cls._get_next_task_sql_and_data()
        return cls.select(connection, sql=sql, data=data)

    @classmethod
    def next_cron_task(cls, connection):
        """Returns a crontask instance which is on due.
        """
        sql, data = cls._get_next_task_sql_and_data()
        sql = f"{sql} AND crontab <> ''"
        return cls.select(connection, sql=sql, data=data)

    @classmethod
    def delete_crontasks(cls, connection):
        """Delete all task which are cron-tasks.
        """
        sql = f"DELETE FROM {cls.table_name} WHERE crontab <> ''"
        connection.run(sql)

    @classmethod
    def change_status(cls, connection, prev_status, new_status):
        """Change the status of a task from a given one to a new one.
        """
        sql = f"""UPDATE {cls.table_name} SET status = :new_status
                  WHERE status == :prev_status"""
        data = {"prev_status": prev_status, "new_status": new_status}
        connection.run(sql, data)

    @staticmethod
    def row_factory(cursor, row):
        """
        SQLite factory class to convert a row from a task-table to a
        dictionary.
        """
        function_arguments_column_name = "function_arguments"
        data = {}
        column_names = [entry[0] for entry in cursor.description]
        for name, value in zip(column_names, row):
            if name == function_arguments_column_name:
                args, kwargs = pickle.loads(value)
                data["args"] = args
                data["kwargs"] = kwargs
            else:
                data[name] = value
        return data


class Result(Model):
    """The model to store function result of delayed executed tasks.
    """

    table_name = "result"
    columns = {
        "uuid": "TEXT PRIMARY KEY",
        "status": "INTEGER",
        "function_module": "TEXT",
        "function_name": "TEXT",
        "function_arguments": "BLOB",
        "function_result": "BLOB",
        "error_message": "TEXT",
        "ttl": "datetime"
    }

    def __init__(
        self,
        connection=None,
        func=None,
        args=(),
        kwargs=None,
        function_result=None,
        error_message="",
        uuid="",
        status=TASK_STATUS_WAITING,
        ttl=None,
        function_name = "",
        function_module = "",
        function_arguments = None,
        rowid = None
    ):
        super().__init__(connection=connection)
        self.rowid = rowid
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.uuid = uuid
        self.status = status
        self.function_name = function_name
        self.function_module = function_module
        self.function_arguments = function_arguments
        self.function_result = function_result
        self.error_message = error_message
        self.ttl = ttl if ttl else datetime.datetime.now()

    @property
    def has_error(self):
        """
        Returns True if the error message is not empty. The return value
        is invalid as long as ``is_ready()`` does not return True.
        """
        return bool(self.error_message)

    @property
    def result(self):
        """
        Shortcut for the attribute ``function_result``. The return value is invalid as
        long as ``is_ready()`` does not return True.
        """
        return self.function_result

    def is_ready(self):
        """
        Returns ``True`` if the task has been processed, otherwise ``False``. If
        the task has been processed the result may be available or an
        error-message may be set.

        Note for async frameworks: this is a *blocking call*.
        """
        processed_states = (TASK_STATUS_READY, TASK_STATUS_ERROR)
        if self.status not in processed_states:
            # connect to database for a refresh of the data
            interface = SQLiteInterface()
            result = interface.get_result_by_uuid(uuid=self.uuid)
            if result is not None:
                self.__dict__.update(result.__dict__)
            return self.status in processed_states
        return True

    def store(self):
        """Stores the result as a new entry in the result-table.
        """
        if self.func:
            self.function_module = self.func.__module__
            self.function_name = self.func.__name__
        else:
            self.function_module = ""
            self.function_name = ""
        self.function_arguments = pickle.dumps((self.args, self.kwargs))
        self.function_result = pickle.dumps(self.function_result)
        super().store()

    @classmethod
    def from_registration(cls, func, args, kwargs, uuid="",
                          status=TASK_STATUS_WAITING,
                          function_result=None, error_message=""):
        """
        Return a new instance with the given arguments as attributes.
        """
        return cls(
            func=func, args=args, kwargs=kwargs, uuid=uuid,
            function_result=function_result,status=status,
            error_message=error_message
        )

    @classmethod
    def from_uuid(cls, connection, uuid):
        """
        Returns a Result instance from the database with the given uuid.
        If there is no entry return None.
        """
        sql = cls._get_sql_select()
        sql = f"{sql} WHERE uuid == :uuid"
        data = {"uuid": uuid}
        return cls.select(connection, sql=sql, data=data)

    @classmethod
    def delete_outdated(cls, connection, schedule):
        """
        Delete all result entries where the ttl is <= schedule.
        (ttl is a datetime object and works as a timestamp for storage.)
        """
        sql = f"""DELETE FROM {cls.table_name}
                  WHERE status <> {TASK_STATUS_WAITING}
                  AND ttl <= :ttl"""
        connection.run(sql, {"ttl": schedule})

    @staticmethod
    def row_factory(cursor, row):
        """
        SQLite factory class to convert a row from the result-table to a
        dictionary.
        """
        data = {}
        column_names = [entry[0] for entry in cursor.description]
        for name, value in zip(column_names, row):
            if name in ("function_arguments", "function_result"):
                data[name] = pickle.loads(value)
            else:
                data[name] = value
        return data


class Settings(Model):
    """
    Model with a single entry in the database storing the settings.
    """

    table_name = "settings"
    columns = {
        "max_workers": "INTEGER",
        "running_workers": "INTEGER",
        "monitor_lock": "INTEGER",
        "autocron_lock": "INTEGER",
        "monitor_idle_time": "INTEGER",
        "worker_idle_time": "INTEGER",
        "worker_pids": "TEXT",
        "result_ttl": "INTEGER"
    }

    def __init__(self, connection=None, data=None):
        """
        Initializes settings with the values from data. data must be a
        dictionary with key value pairs according the `columns` class
        attribute. If data is not given the
        SETTINGS_DEFAULT_DATA-dictionary is used to populate the
        settings data.
        """
        super().__init__(connection)
        data = data if data else SETTINGS_DEFAULT_DATA
        self.__dict__.update(data)

    def __repr__(self):
        """Self representation used by the admin-tool.
        """
        width = len(max(self.columns, key=len))
        attributes = []
        for key in self.columns:
            value = self.__dict__[key]
            attributes.append(f"{key:<{width}}: {value}")
        return "\n".join(attributes)

    @classmethod
    def read(cls, connection):
        """
        Returns a settings instance with data read from the database. If
        there is no settings entry in the table returns None.
        """
        entries = cls.select_all(connection)
        if not entries:
            return None
        return entries[0]

    @staticmethod
    def row_factory(cursor, row):
        """
        SQLite factory function to convert a row from a settings-table
        to a dictionary.
        """
        column_names = [entry[0] for entry in cursor.description]
        data = {name: bool(value) if name in BOOLEAN_SETTINGS else value
                for name, value in zip(column_names, row)}
        return data


class TaskRegistrator:
    """
    Handles the task registration in a separate thread so that
    registration is a non-blocking operation.
    """

    def __init__(self, interface):
        self.interface = interface
        self.task_queue = queue.Queue()
        self.exit_event = threading.Event()
        self.registration_thread = None

    def register(self, func, args=(), kwargs=None,
                       crontab="", uuid="", schedule=None):
        """
        Register a task for later processing. Arguments are the same as
        for `SQLiteInterface.register_task()` which is called from a
        seperate thread.
        """
        if kwargs is None:
            kwargs = {}
        self.task_queue.put({
            "func": func,
            "schedule": schedule,
            "crontab": crontab,
            "uuid": uuid,
            "args": args,
            "kwargs": kwargs,
        })

    def _process_queue(self):
        """
        Register task in a separate thread taking the tasks from a
        task_queue.
        """
        while True:
            try:
                data = self.task_queue.get(
                    timeout=REGISTER_BACKGROUND_TASK_TIMEOUT
                )
            except queue.Empty:
                # check for exit_event on empty queue so the queue items
                # can get handled before terminating the thread
                if self.exit_event.is_set():
                    break
            else:
                self.interface.register_task(**data)

    def start(self):
        """
        Start processing the queue in a seperate thread.
        """
        # don't start multiple threads
        if self.registration_thread is None:
            self.registration_thread = threading.Thread(
                target=self._process_queue
            )
            self.registration_thread.start()

    def stop(self):
        """
        Terminates the running registration thread.
        """
        if self.registration_thread:
            self.exit_event.set()
            self.registration_thread = None


class SQLiteConnection:
    """
    SQLite connection. `run()` can get called as often as required. The
    database keeps connected. Leaving the context will commit and close
    the database connection. In case of an exception during the context
    instead of a commit the connection will do a rollback.
    """

    def __init__(self, db_name, row_factory=None, exclusive=False):
        self.row_factory = row_factory
        self.db_name = db_name
        self.connection = None
        self.exclusive = exclusive

    def __enter__(self):
        self.connection = sqlite3.connect(
            database=self.db_name,
            detect_types=sqlite3.PARSE_DECLTYPES
        )
        if self.row_factory:
            self.connection.row_factory = self.row_factory
        if self.exclusive:
            self.connection.execute(SQLITE_EXCLUSIVE_ACCESS)
        return self

    def __exit__(self, *args):
        if any(args):
            # there was an exception:
            self.connection.rollback()
        else:
            self.connection.commit()
        self.connection.close()

    def run(self, command, parameters=(), many=None):
        """
        Runs an sql command with the given parameters. If the command
        supports qmark style, the parameters must be a tuple with the
        parameters in the proper order. If many is True, the parameters
        must be a sequence of tuples.
        If the command supports named style, parameters should be a
        dictionary. If many is True, parameters should be a sequence of
        dicts.
        Returns the result of the given command.
        """
        if parameters and many is None:
            # try to find out whether it is many or not:
            if isinstance(parameters, (tuple, list)):
                # could be qmark parameters or a sequence of tuples or dicts
                # take the first element of the sequence, if this is another
                # sequence or dict then these are parameters for an
                # executemany() command, else just execute()
                many = isinstance(parameters[0], (tuple, list, dict))
        if many:
            return self.connection.executemany(command, parameters)
        return self.connection.execute(command, parameters)


Connection = SQLiteConnection


class SQLiteInterface:
    """
    SQLite interface for application specific operations.
    This class is a Singleton.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        # run __init__ just once
        if self.__dict__:
            return
        self._db_name = None
        self._result_ttl = None
        # accept_registrations will get set to False
        # by the worker processes to not register callables by the workers
        self.accept_registrations = True
        # these attributes are set later by reading the settings:
        self.autocron_lock = None
        self.monitor_lock = None
        self.monitor_idle_time = None
        self.max_workers = None
        self.worker_idle_time = None
        # the registrator for non blocking registration:
        self.registrator = TaskRegistrator(self)

    @property
    def db_name(self):
        """Return hidden _db_name (because of the setter)"""
        return self._db_name

    @db_name.setter
    def db_name(self, db_name):
        """
        Expand the given database filename to the storage-location path
        if the filename is not an absolute path. In the latter case the
        absolute path will get used (and must exist). If db_name is None
        set None directly.
        """
        if db_name is None:
            self._db_name = None
        else:
            path = pathlib.Path(db_name)
            if not path.is_absolute():
                try:
                    path = pathlib.Path.home() / DEFAULT_STORAGE / path.name
                except RuntimeError:
                    # no home directory found
                    path = pathlib.Path.cwd() / db_name
            self._db_name = path

    @property
    def result_ttl(self):
        """
        Returns the new ttl as a datetime instance with an offset of now.
        """
        return datetime.datetime.now() + self._result_ttl

    @result_ttl.setter
    def result_ttl(self, value=SETTINGS_DEFAULT_RESULT_TTL):
        """
        Set the ttl as timedelta in seconds. This is the lifetime from
        an offset to the future.
        """
        self._result_ttl = datetime.timedelta(seconds=value)

    @db_access
    def init_database(self, db_name):
        """
        Set the database name and set up initial data.
        """
        if not self.db_name:
            self.db_name = db_name
            with Connection(self.db_name, exclusive=True) as conn:
                Task.create_table(conn)
                Result.create_table(conn)
                Settings.create_table(conn)

                # try to read the settings. If this fails create the first
                # (and only) settings dataset with the default values:
                settings = Settings.read(conn)
                if settings is None:
                    settings = Settings(conn)  # this set the defaults
                    settings.store()  # store defaults in the database

                # set attributes from settings that don't change during runtime:
                self.autocron_lock = settings.autocron_lock
                self.monitor_lock = settings.monitor_lock
                self.monitor_idle_time = settings.monitor_idle_time
                self.max_workers = settings.max_workers
                self.worker_idle_time = settings.worker_idle_time
                self.result_ttl = settings.result_ttl

    @db_access
    def register_task(self, func, schedule=None, crontab="", uuid="",
                      args=(), kwargs=None):
        """
        Store a callable in the task-table of the database. If the
        callable is a delayed task with a potential result create also a
        corresponding entry in the result table. If the callable is a
        crontask, check whether the task is already registered and don't
        register the callable again.
        """
        if self.accept_registrations:
            if not schedule:
                schedule = datetime.datetime.now()
            if kwargs is None:
                kwargs = {}
            with Connection(self.db_name, exclusive=True) as conn:
                if crontab and Task.get_by_function_name(func, conn):
                    # don't register a crontab twice:
                    return
                task = Task(
                    connection=conn,
                    func=func,
                    schedule=schedule,
                    crontab=crontab,
                    uuid=uuid,
                    args=args,
                    kwargs=kwargs
                )
                task.store()

                # if a uuid is given it is a delayed function that
                # may return a result:
                if uuid:
                    result = Result(
                        connection=conn,
                        func=func,
                        args=args,
                        kwargs=kwargs,
                        uuid=uuid,
                        ttl=self.result_ttl
                    )
                    result.store()

    @db_access
    def get_next_task(self):
        """
        Returns the next task on due with crontasks first or None if
        there is not task on due. If a task is returned the status is
        set to TASK_STATUS_PROCESSING first.
        """
        with Connection(self.db_name, exclusive=True) as conn:
            task = Task.next_cron_task(conn) or Task.next_task(conn)
            if task:
                task.status = TASK_STATUS_PROCESSING
                task.update()
        return task

    @db_access
    def update_task_schedule(self, task, schedule):
        """Updates the schedule of the given task.
        """
        with Connection(self.db_name) as conn:
            task.connection = conn
            task.schedule = schedule
            task.status = TASK_STATUS_WAITING
            task.update()

    @db_access
    def count_tasks(self):
        """Return the number of entries in the task-table.
        """
        with Connection(self.db_name) as conn:
            return Task.count_rows(conn)

    @db_access
    def get_tasks(self):
        """Return a list of all tasks.
        """
        with Connection(self.db_name) as conn:
            return Task.select_all(conn)

    @db_access
    def delete_task(self, task):
        """Delete the task which may not have a valid connection-attribute.
        """
        # solution: inject a valid connection
        with Connection(self.db_name) as conn:
            task.connection = conn
            task.delete()

    @db_access
    def get_result_by_uuid(self, uuid):
        """
        Return a Result instance from the database identified by the uuid.
        """
        with Connection(self.db_name) as conn:
            return Result.from_uuid(connection=conn, uuid=uuid)

    @db_access
    def count_results(self):
        """Return the number of entries in the task-table.
        """
        with Connection(self.db_name) as conn:
            return Result.count_rows(conn)

    @db_access
    def update_result(
        self,
        uuid,
        result=None,
        error_message="",
        ttl=None
    ):
        """
        Updates the result with the uuid with the values of the
        arguments result and error_message.
        """
        function_result = pickle.dumps(result)
        ttl = ttl if ttl else self.result_ttl
        status = TASK_STATUS_ERROR if error_message else TASK_STATUS_READY
        with Connection(self.db_name) as conn:
            result = Result.from_uuid(conn, uuid=uuid)
            result.function_result = function_result
            result.function_arguments = pickle.dumps(result.function_arguments)
            result.error_message = error_message
            result.status = status
            result.ttl = ttl
            result.update()

    @db_access
    def delete_outdated_results(self):
        """Delete all resuts with a ttl <= now.
        """
        with Connection(self.db_name) as conn:
            Result.delete_outdated(conn, datetime.datetime.now())

    @db_access
    def increment_running_workers(self, pid):
        """
        Add the pid to the worker pid-list and increase the running
        worker num by 1.
        """
        with Connection(self.db_name) as conn:
            settings = Settings.read(connection=conn)
            settings.worker_pids = f"{settings.worker_pids},{pid}".lstrip(",")
            settings.running_workers += 1
            settings.update()

    @db_access
    def decrement_running_workers(self, pid):
        """
        Delete the pid from the worker_pids list and decrement the
        running_workers counter.
        """
        with Connection(self.db_name) as conn:
            settings = Settings.read(connection=conn)
            pids = settings.worker_pids.split(",")
            try:
                pids.remove(str(pid))
            except ValueError:
                # pid not in list: ignore
                pass
            else:
                settings.worker_pids = ",".join(pids)
                settings.running_workers = len(pids)
                settings.update()

    @db_access
    def is_worker_pid(self, pid):
        """Check whether the provided pid is one of the worker pids.
        """
        with Connection(self.db_name) as conn:
            settings = Settings.read(connection=conn)
        pids = (int(p) for p in settings.worker_pids.split(",") if p)
        return pid in pids

    @db_access
    def acquire_monitor_lock(self):
        """
        Tries to acquire the monitor-lock flag: if the flag is set to
        False, then set it to True. Otherwise keep the flag as is.
        Return True if the flag has been set to True, otherwise return
        False.
        """
        with Connection(self.db_name, exclusive=True) as conn:
            settings = Settings.read(connection=conn)
            if not settings.monitor_lock:
                settings.monitor_lock = True
                settings.update()
                return True
            return False

    @db_access
    def get_settings(self):
        """Returns the settings dataset.
        """
        with Connection(self.db_name) as conn:
            return Settings.read(connection=conn)

    @db_access
    def update_settings(self, settings):
        """Updates the settings dataset.
        """
        with Connection(self.db_name) as conn:
            settings.connection = conn
            settings.update()

    @db_access
    def tear_down_database(self):
        """
        Reset all settings here so that the workers don't have to access
        the database again on shutdown.
        """
        # gets called from the engine in case the interface
        # is the worker_master
        with Connection(self.db_name, exclusive=True) as conn:
            settings = Settings.read(conn)
            settings.monitor_lock = False
            settings.running_workers = 0
            settings.worker_pids = ""
            settings.update()
            Task.delete_crontasks(conn)
            # reset the status of unfinished tasks from the
            # last run to handle them again:
            Task.change_status(
                conn,
                prev_status=TASK_STATUS_PROCESSING,
                new_status=TASK_STATUS_WAITING
            )



"""
test_engine.py

tests for the engine.
"""

import pathlib
import subprocess
import time
import warnings

import pytest

from autocron import engine
from autocron import sqlite_interface


TEST_DB_NAME = "test.db"


# all labels starting with 'tst_' are test-functions:

def tst_add():
    pass

def tst_cron():
    pass

def tst_cron2():
    pass


@pytest.fixture
def interface():
    """
    Returns a new database instance.
    """
    # set class attribute to None to not return a singleton
    sqlite_interface.SQLiteInterface._instance = None
    interface = sqlite_interface.SQLiteInterface()
    yield interface
    if interface.db_name:
        pathlib.Path(interface.db_name).unlink(missing_ok=True)


def test_start_subprocess():
    """
    Start and stop a subprocess.
    """
    # Needs interface.db_name as argument but makes no use of it.
    process = engine.start_subprocess("unused_db_name")
    assert isinstance(process, subprocess.Popen) is True
    assert process.poll() is None
    process.terminate()
    time.sleep(0.02)  # give process some time to terminate
    assert process.poll() is not None


def test_start_and_stop_workerprocess():
    """
    Test to start and terminate a subprocess.
    """
    warnings.simplefilter("ignore", ResourceWarning)
    process = engine.start_subprocess(None)
    assert process.poll() is None  # subprocess runs
    process.terminate()
    time.sleep(0.2)
    assert process.poll() is not None


def test_delete_crontasks_on_shutdown(interface):
    """
    Scenario: put a cron-task in the database. Then close the
    database (to simulate a shutdown of the application). The
    cron-task should be deleted.
    """
    interface.init_database(db_name=TEST_DB_NAME)

    # put two crontasks and a normal task in the database:
    interface.register_task(tst_add)
    interface.register_task(tst_cron, crontab="* * * * *")
    interface.register_task(tst_cron2, crontab="* * * * *")
    entries = interface.count_tasks()
    assert entries == 3

    # simulate an engine in running mode and stop it:
    engine_ = engine.Engine(interface=interface)
    engine_.monitor_thread = True
    engine_.stop()

    # the crontasks should now be deleted:
    entries = interface.count_tasks()
    assert entries == 1

    # and the remaining task should be the add-task:
    task = interface.get_tasks()[0]
    assert bool(task.crontab) is False




"""
test_worker.py

test worker functionality. Start and stop of worker process is testet in
test_engine.py
"""
0
import pathlib

import pytest

from autocron import sqlite_interface
from autocron import worker


TEST_DB_NAME = "test.db"


def tst_add(a, b):
    return a + b


@pytest.fixture
def interface():
    """
    Returns a new initialised database instance.
    """
    # set class attribute to None to not return a singleton
    sqlite_interface.SQLiteInterface._instance = None
    interface = sqlite_interface.SQLiteInterface()
    interface.init_database(db_name=TEST_DB_NAME)
    yield interface
    if interface.db_name is not None:
        pathlib.Path(interface.db_name).unlink(missing_ok=True)


def test_init_worker(interface):
    """
    Should run without error after refactoring and because the interface
    is a singleton and set during the __init__ method.
    """
    assert interface.accept_registrations is True
    worker_ = worker.Worker(TEST_DB_NAME)
    assert worker_.interface is interface


def test_handle_delayed_task(interface):
    """
    Handle a simple task and check for the correct processing.
    """
    # register a delayed task (the uuid must just be a unique string)
    uuid_ = "testid"
    interface.register_task(func=tst_add, args=(40, 2), uuid=uuid_)

    # check for table entries:
    assert interface.count_tasks() == 1
    assert interface.count_results() == 1

    # create a worker and handle the task:
    worker_ = worker.Worker(TEST_DB_NAME)
    has_processed_a_task = worker_.handle_task()
    assert has_processed_a_task is True

    # check for table entries (the task should have been deleted):
    assert interface.count_tasks() == 0
    assert interface.count_results() == 1

    # get the result by the known uuid:
    result = interface.get_result_by_uuid(uuid=uuid_)
    assert result.is_ready() is True
    assert result.function_result == 42




from datetime import datetime as dt

import pytest

from autocron.schedule import (
    get_cron_parts,
    get_next_value,
    get_numeric_sequence,
    get_weekday,
    CronScheduler,
)


@pytest.mark.parametrize(
    'value, values, expected_result', [
        (2, [5, 10, 15], 5),
        (5, [5, 10, 15], 10),
        (10, [5, 10, 15], 15),
        (15, [5, 10, 15], None),
        (20, [5, 10, 15], None),
        (2, [7], 7),
        (7, [7], None),
        (9, [7], None),
    ])
def test_get_next_value(value, values, expected_result):
    """
    Check to get the next element from a sequence of values larger than
    a given value. If there is no larger value, None gets returned
    """
    result = get_next_value(value, values)
    assert result == expected_result


@pytest.mark.parametrize(
    "pattern, min_value, max_value, expected_result", [
        ("*", 0, 5, list(range(6))),
        ("*/2", 0, 6, list(range(0, 7, 2))),
        ("*/15", 0, 59, list(range(0, 60, 15))),
        ("12,5,30", 0, 59, [5, 12, 30]),
        ("10-20", 0, 59, list(range(10, 21))),
        ("2,4-8,22", 0, 59, [2]+list(range(4, 9))+[22]),
        ("20-30,8,6,4,10-19",
            0, 59, list(range(4, 9, 2)) + list(range(10, 31))),
        ("*", 1, 30, list(range(1, 31))),
        ("*/2", 0, 10, list(range(0, 11, 2))),
        ("*/2", 1, 10, list(range(1, 11, 2))),
    ]
)
def test_get_numeric_sequence(pattern, min_value, max_value, expected_result):
    """
    Test to convert a pattern like "*", "*/n" etc. to a sequence of
    numeric values.
    """
    result = get_numeric_sequence(pattern, min_value, max_value)
    assert result == expected_result


def test_get_cron_parts():
    """
    Test the crontab parsing into list of values.
    """
    crontab = "2,3-5 * 2-4 */4 */2"
    cp = get_cron_parts(crontab)
    assert cp.minutes == [2, 3, 4, 5]
    assert cp.hours == list(range(24))
    assert cp.days == [2, 3, 4]
    assert cp.months == [1, 5, 9]
    assert cp.days_of_week == [0, 2, 4, 6]


def test_cronscheduler_init():
    """
    Test if the CronScheduler converts the init-data according to
    test_get_cron_parts in case that the keyword-arguments minutes,
    hours etc. are given.
    """
    def check_cp():
        cp = cs.cron_parts
        assert cp.minutes == [2, 3, 4, 5]
        assert cp.hours == list(range(24))
        assert cp.days == [2, 3, 4]
        assert cp.months == [1, 5, 9]
        assert cp.days_of_week == [0, 2, 4, 6]

    # test with crontab
    crontab = "2,3-5 * 2-4 */4 */2"
    cs = CronScheduler(crontab)
    check_cp()

    # test with keyword arguments
    cs = CronScheduler(
        minutes=[2, 3, 4, 5],
        days=[2, 3, 4],
        months=[1, 5, 9],
        days_of_week=[0, 2, 4, 6]
    )
    check_cp()


@pytest.mark.parametrize(
    'schedule, expected_result', [
        (dt(2024, 2, 12), 1),  # this is a monday
        (dt(2024, 2, 11), 0),  # this is a sunday
    ]
)
def test_get_weekday(schedule, expected_result):
    result = get_weekday(schedule=schedule)
    assert result == expected_result


@pytest.mark.parametrize(
    'crontab, minute, expected_result', [
        ("10,20 * * * *", 5, 10),
        ("10,20 * * * *", 15, 20),
        ("10,20 * * * *", 20, None),
        ("10,20 * * * *", 25, None),
    ]
)
def test_get_next_minute(crontab, minute, expected_result):
    cs = CronScheduler(crontab)
    result = cs.get_next_minute(minute)
    assert result == expected_result


@pytest.mark.parametrize(
    'crontab, hour, expected_result', [
        ("* 10,20 * * *", 5, 10),
        ("* 10,20 * * *", 15, 20),
        ("* 10,20 * * *", 20, None),
        ("* 10,20 * * *", 23, None),
    ]
)
def test_get_next_hour(crontab, hour, expected_result):
    cs = CronScheduler(crontab)
    result = cs.get_next_hour(hour)
    assert result == expected_result


@pytest.mark.parametrize(
    'crontab, month, expected_result', [
        ("* * * * *", 5, 6),
        ("* * * 7 *", 5, 7),
        ("* * * 7,12 *", 7, 12),
        ("* * * 7 *", 7, None),
        ("* * * */2 *", 3, 5),
        ("* * * */2 *", 11, None),
    ]
)
def test_get_next_month(crontab, month, expected_result):
    cs = CronScheduler(crontab)
    result = cs.get_next_month(month)
    assert result == expected_result


@pytest.mark.parametrize(
    'crontab, year, month, strict_mode, expected_result', [
        ("* * * * *", 2024, 2, False, 1),
        ("* * 10 * *", 2024, 2, False, 10),
        ("* * 10 * 1", 2024, 2, False, 5),
        ("* * 10 * 6", 2024, 2, True, 10),
        ("* * 10 * 5", 2024, 2, True, None),
    ]
)
def test_get_first_day(crontab, year, month, strict_mode, expected_result):
    cs = CronScheduler(crontab, strict_mode=strict_mode)
    result = cs.get_first_day(year, month)
    assert result == expected_result


@pytest.mark.parametrize(
    'crontab, year, month, day, strict_mode, expected_result', [
        ("* * * * *", 2024, 2, 2, False, 3),
        ("* * * * *", 2024, 2, 28, False, 29),
        ("* * * * *", 2025, 2, 28, False, None),
        ("* * 10,20 * *", 2024, 2, 5, False, 10),
        ("* * 10,20 * *", 2024, 2, 10, False, 20),
        ("* * 10,20 * *", 2024, 2, 22, False, None),
        ("* * 10,20 * 1", 2024, 2, 28, False, None),
        ("* * 10,20 * 2,5", 2024, 2, 2, False, 6),
        ("* * 10,20 * 2,5", 2024, 2, 14, False, 16),
        ("* * 10,20 * 2,5", 2024, 2, 23, False, 27),
        ("* * 10,20 * 2,5", 2024, 2, 2, True, 20),
        ("* * 10,21 * 3", 2024, 2, 2, True, 21),
        ("* * 10,22 * 3", 2024, 2, 2, True, None),
    ]
)
def test_get_next_day(crontab, year, month, day, strict_mode, expected_result):
    cs = CronScheduler(crontab, strict_mode=strict_mode)
    result = cs.get_next_day(year, month, day)
    assert result == expected_result


@pytest.mark.parametrize(
    'crontab, previous_schedule, strict_mode, expected_result', [
        ("* * * * *", dt(2024, 2, 8, 10, 0), False, dt(2024, 2, 8, 10, 1)),
        ("*/5 * * * *", dt(2024, 2, 8, 10, 24), False, dt(2024, 2, 8, 10, 25)),
        ("*/5 * * * *", dt(2024, 2, 8, 10, 25), False, dt(2024, 2, 8, 10, 30)),
        ("0,30 * * * *", dt(2024, 2, 8, 10, 25), False, dt(2024, 2, 8, 10, 30)),
        ("0,30 * * * *", dt(2024, 2, 8, 10, 30), False, dt(2024, 2, 8, 11, 0)),
        ("0,30 5,17 * * *", dt(2024, 2, 8, 10, 30), False, dt(2024, 2, 8, 17, 0)),
        ("0,30 5,17 * * *", dt(2024, 2, 8, 17, 30), False, dt(2024, 2, 9, 5, 0)),
        ("30 13 * * 5", dt(2024, 2, 8, 17, 30), False, dt(2024, 2, 9, 13, 30)),
        ("30 13 * * 5", dt(2024, 2, 9, 13, 30), False, dt(2024, 2, 10, 13, 30)),
        ("30 13 * * 5", dt(2024, 2, 9, 13, 30), True, dt(2024, 2, 16, 13, 30)),
        ("30 13 * * 5", dt(2024, 2, 16, 13, 30), True, dt(2024, 2, 23, 13, 30)),
        ("30 13 * * 5", dt(2024, 2, 23, 13, 30), True, dt(2024, 3, 1, 13, 30)),
        ("30 13 29 2 *", dt(2024, 2, 27, 13, 30), False, dt(2024, 2, 29, 13, 30)),
        ("30 13 29 2 *", dt(2024, 2, 29, 13, 30), False, dt(2028, 2, 29, 13, 30)),
        ("30 13 29 2 0", dt(2024, 2, 29, 13, 30), False, dt(2025, 2, 2, 13, 30)),
        ("30 13 29 2 0", dt(2024, 2, 29, 13, 30), True, dt(2032, 2, 29, 13, 30)),
        ("30 13 7 * 3", dt(2024, 2, 7, 13, 30), True, dt(2024, 8, 7, 13, 30)),
        ("30 13 7 2 3", dt(2024, 2, 7, 13, 30), True, dt(2029, 2, 7, 13, 30)),
    ]
)
def test_get_next_schedule(crontab,
                           previous_schedule,
                           strict_mode,
                           expected_result):
    cs = CronScheduler(crontab, strict_mode=strict_mode)
    result = cs.get_next_schedule(previous_schedule)
    assert result == expected_result
